/*
 * radio-sf64pcr.c
 * 
 * Soundforte 64 pcr PCI radio card driver for video4linux
 * (C) 2003 Minos Panagiotis <panosminos@mycosmos.gr>
 *
 * The card uses the TEA5757 chip for the tuner
 * and the FM801AU chip for the PCI controller.
 *
 * Some notes on hardware.
 *
 * The TEA5757 chip shift register is controlled by 4 lines (Write
 * Enable, Clock, Data & Status) that are connected in the GPIO of the
 * PCI Controller (FM801AU).
 *
 * GPIO Bit	Pin of TEA5757	input / output
 *	1	clk		always output
 *	2	we(inverted)	always output
 *	3	data		input / output
 *	4	mo/st		always input
 */


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/videodev.h>

#include <asm/io.h>

#define DRV_NAME        "radio-sf64pcr"
#define DRV_VERSION     "0.1"
#define DRV_RELDATE     "20/05/2003"
#define DRV_AUTHOR	"Panos Minos, <panosminos@mycosmos.gr>"

#ifndef PCI_VENDOR_ID_FORTEMEDIA
#define PCI_VENDOR_ID_FORTEMEDIA        0x1319
#endif

#ifndef PCI_DEVICE_ID_FORTEMEDIA_FM801
#define PCI_DEVICE_ID_FORTEMEDIA_FM801  0x0801
#endif

#define FREQ_IF         171200  /* 10.7*16000   */
#define FREQ_STEP       200     /* 12.5*16      */
#define VID_HARDWARE_SF64PCR 10000
#define FREQ_LO (87.5*16000)
#define FREQ_HI (108*16000)

struct sf64pcr_card {
	u32 iobase;
	u32 length;
	int muted;
	int stereo;
	int is_stereo;
	int is_tuned;
	unsigned long freq;
};
static void sf64pcr_set_frequency(struct sf64pcr_card *card,
					unsigned long freq);
static unsigned long sf64pcr_get_frequency(struct sf64pcr_card *card);
/****************************************************************************/

static struct sf64pcr_card  sf64pcr_card = { 0,0,0,1,0,0,0 };
static const char version[] = DRV_NAME ".c:v" DRV_VERSION " "
DRV_RELDATE " " DRV_AUTHOR "\n";

static int is_stereo(struct sf64pcr_card *card)
{

	sf64pcr_get_frequency(card);
	return card->is_stereo;
}

static int is_tuned(struct sf64pcr_card *card)
{
	sf64pcr_get_frequency(card);
	return card->is_tuned;
}

static void sf64pcr_mute(struct sf64pcr_card *card)
{
	outw(0xf800, card->iobase+0x52);
	card->muted = 1;

}

static void sf64pcr_unmute(struct sf64pcr_card *card)
{
	outw(0xf802, card->iobase+0x52);
	card->muted = 0;

}

static unsigned long read_shift_register(struct sf64pcr_card *card)
{
	unsigned long data = 0UL;
	unsigned short ioport = card->iobase + 0x52;
	int c;

	outw(0xfc02,ioport);
	udelay(4);
	for (c = 23; c >= 0 ; c--) {
		data <<= 1;
		outw(0xfc03,ioport);
		udelay(4);

		if (c == 0) {
			if ((inb(ioport) & 0x08) == 0)
				card->is_tuned = 1;
			else
				card->is_tuned = 0;
		}
		outw(0xfc02,ioport);

		if (inw(ioport) & 0x04)
			data |= 1;
		if (c == 0) {
			if ((inb(ioport) & 0x08) == 0)
				card->is_stereo = 1;
			else
				card->is_stereo = 0;
		}

			udelay(4);
	}
	return data;

}

static void write_shift_register(struct sf64pcr_card *card, unsigned long data)
{
	int c;
	unsigned int ioport = card->iobase + 0x52;	

	outw(0xf800,ioport); udelay(4);

    for (c = 24; c >= 0; c--) {
        if (data & (1 << c) ) {
            outw(0xf804,ioport); udelay(4);
            outw(0xf805,ioport); udelay(4);
            outw(0xf804,ioport); udelay(4);
        }
        else {
            outw(0xf800,ioport); udelay(4);
            outw(0xf801,ioport); udelay(4);
            outw(0xf800, ioport); udelay(4);
        }
    }
	if(!card->muted)
		outw(0xf802,ioport);
}

static unsigned long sf64pcr_get_frequency(struct sf64pcr_card *card)
{
        unsigned long data;
        data = read_shift_register(card);
        data &= 0x0003FFF;
        data = data * FREQ_STEP - FREQ_IF;
	card->freq = data;
        return data;
}

static void sf64pcr_set_frequency(struct sf64pcr_card *card, unsigned
	long freq)
{
	card->freq = freq;
	freq += 10700*16;
	freq /= FREQ_STEP;
	freq &= 0x0003FFF;
	if(card->stereo == 0) 
		freq |= 0x00400000;
	
	write_shift_register(card, freq);
}

static void sf64pcr_stereo_on(struct sf64pcr_card *card)
{
        card->stereo = 1;
        sf64pcr_set_frequency(card,card->freq);
}

static void sf64pcr_stereo_off(struct sf64pcr_card *card)
{
        card->stereo = 0;
        sf64pcr_set_frequency(card,card->freq);
}
/****************************************************************************/

static int users = 0;

static int sf64pcr_open(struct video_device *dev, int flags)
 {
     if(users)
         return -EBUSY;
     users++;
     MOD_INC_USE_COUNT;
     return 0;
 }

static void sf64pcr_close(struct video_device *dev)
 {
     users --;
     MOD_DEC_USE_COUNT;
 }

static int sf64pcr_ioctl(struct video_device *dev, unsigned int cmd,
                         void *arg)
{
	switch (cmd) {
		case VIDIOCGCAP:
		{
			struct video_capability v;
			v.type = VID_TYPE_TUNER;
			v.channels = 1;
			v.audios = 1;
			v.maxwidth = 0;
			v.maxheight = 0;
			v.minwidth = 0;
			v.minheight = 0;
			strcpy(v.name, "SF 64 PCR PCI Radio");

			if (copy_to_user(arg, &v, sizeof(v)))
				return -EFAULT;
			return 0;
		}
		case VIDIOCGTUNER:
		{
			struct video_tuner v;

			if (copy_from_user(&v, arg, sizeof(v))!=0)
				return -EFAULT;
			if (v.tuner)
				return -EINVAL;
			v.rangelow = FREQ_LO;
			v.rangehigh = FREQ_HI;
			v.flags = VIDEO_TUNER_LOW;
			
			if(is_stereo(&sf64pcr_card))
				v.flags |= VIDEO_TUNER_STEREO_ON;
			v.mode = VIDEO_MODE_AUTO;
			
			if(is_tuned(&sf64pcr_card))
				v.signal = 0xffff;
			else
				v.signal = 0;
			strcpy(v.name, "FM");

			if (copy_to_user(arg, &v, sizeof(v))!=0)
				return -EFAULT;
			return 0;
		}
		case VIDIOCSTUNER:
		{
			struct video_tuner v;

			if (copy_from_user(&v, arg, sizeof(v)))
				return -EFAULT;
			if (v.tuner)
				return -EINVAL;
			return 0;
		}
		case VIDIOCGFREQ:
		{
		unsigned long f;

		f = sf64pcr_get_frequency(&sf64pcr_card);

		if (copy_to_user(arg, &f, sizeof(f)))
			return -EFAULT;
		return 0;
		}
		case VIDIOCSFREQ:
		{

		unsigned long f;

		if (copy_from_user(&f, arg, sizeof(f)))
			return -EFAULT;
		if (f < FREQ_LO || f > FREQ_HI)
			return -EINVAL;

		sf64pcr_set_frequency(&sf64pcr_card,f);
			return 0;
		}
		case VIDIOCGAUDIO:
		{
			struct video_audio v;


			memset(&v,0, sizeof(v));
			if(copy_from_user(&v, arg, sizeof(v)))
				return -EFAULT;
			strcpy(v.name, "Radio");
			v.flags = VIDEO_AUDIO_MUTABLE;
			if(sf64pcr_card.muted)
				v.flags |= VIDEO_AUDIO_MUTE;
			if(is_stereo(&sf64pcr_card))
				v.mode = VIDEO_SOUND_STEREO;
			else
				v.mode = VIDEO_SOUND_MONO;
			if (copy_to_user(arg, &v, sizeof(v)))
				return -EFAULT;
			return 0;
		}
		case VIDIOCSAUDIO:
		{
			struct video_audio v;
			if (copy_from_user(&v, arg, sizeof(v)))
				return -EFAULT;
//			if (v.audio)
//				return -EINVAL;
			if (v.flags & VIDEO_AUDIO_MUTE)
				sf64pcr_mute(&sf64pcr_card);
			else
				sf64pcr_unmute(&sf64pcr_card);
			if(v.flags & VIDEO_SOUND_MONO)
				sf64pcr_stereo_off(&sf64pcr_card);
			if(v.flags & VIDEO_SOUND_STEREO)
				sf64pcr_stereo_on(&sf64pcr_card);
			return 0;
		}
		case VIDIOCGUNIT:
		{
			struct video_unit v;

			v.video = VIDEO_NO_UNIT;
			v.vbi = VIDEO_NO_UNIT;
			v.radio = dev->minor;
			v.audio = 0;
			v.teletext = VIDEO_NO_UNIT;
			if (copy_to_user(arg, &v, sizeof(v)))
				return -EFAULT;
			return 0;
		}
		default:

		return -ENOIOCTLCMD;
	}
}

static struct video_device sf64pcr_radio = {
        owner:          THIS_MODULE,
        name:           "SoundForte 64 PCR Radio Tuner",
        type:           VID_TYPE_TUNER,
        hardware:       VID_HARDWARE_SF64PCR,
        open:           sf64pcr_open,
        close:          sf64pcr_close,
        ioctl:          sf64pcr_ioctl,
};

/****************************************************************************/

static int sf64pcr_found = 0;

/*
 *	This function doesn't support 2 or more cards,
 *	after finding one card will not search
 *	for another.
 */

static int sf64pcr_probe(struct pci_dev *pci_dev,
		 	const struct pci_device_id *pci_id )
{

	if (sf64pcr_found)
		return 0;	

	if(pci_enable_device(pci_dev))
		return -ENODEV;

        sf64pcr_card.iobase = pci_resource_start( pci_dev, 0 );
        sf64pcr_card.length = pci_resource_len( pci_dev, 0 );
	printk(KERN_INFO "%s",version);
	printk(KERN_INFO "sf64-pcr: found radio card at io=%#x.\n",
		sf64pcr_card.iobase);
	
        if(request_region(sf64pcr_card.iobase, sf64pcr_card.length,
		"SF64PCR") == NULL )
	{
                printk(KERN_ERR "sf64-pcr: i/o port already in use\n");
                return -ENODEV;
	}

	if (video_register_device(&sf64pcr_radio, VFL_TYPE_RADIO, -1) == -1)
	{
		release_region(sf64pcr_card.iobase,sf64pcr_card.length );
		printk(KERN_ERR "sf64-pcr: could not register video device\n");
		return -ENODEV;
	}
	sf64pcr_found++;
	return 0;
}


static void sf64pcr_remove( struct pci_dev *pci_dev )
{

	sf64pcr_mute(&sf64pcr_card);

	video_unregister_device(&sf64pcr_radio);
	release_region(sf64pcr_card.iobase,sf64pcr_card.length );

}


static struct pci_device_id sf64pcr_id[] = {
        { PCI_VENDOR_ID_FORTEMEDIA,PCI_DEVICE_ID_FORTEMEDIA_FM801,
          PCI_ANY_ID,PCI_ANY_ID, 0, 0, 0 },
        { 0 }
};
MODULE_DEVICE_TABLE( pci, sf64pcr_id );


static struct pci_driver sf64pcr_pci_driver = {
	name:		"radio-sf64pcr",
	id_table:	sf64pcr_id,
	probe:		sf64pcr_probe,
	remove:		__devexit_p(sf64pcr_remove),
};


static int __init sf64pcr_init_module(void)
{
    return pci_module_init(&sf64pcr_pci_driver);
}


static void __exit sf64pcr_cleanup_module(void)
{
    pci_unregister_driver(&sf64pcr_pci_driver);
}


module_init(sf64pcr_init_module);
module_exit(sf64pcr_cleanup_module);


MODULE_LICENSE("GPL");
MODULE_AUTHOR(DRV_AUTHOR);
MODULE_DESCRIPTION("A driver for the SoundForte 64 PCR Radio Tuner.");
